/*
  ceil_clean.c

  Useful functions for accessing ceilometer data files
  All use the prefix LC, which stands for lidar ceilometer

  R J Hogan, 10 May 1999
  
  8 Oct 2000: Fairly substantial modifications to cleanup function
*/

#include <math.h>
#include "ceil.h"

/* Remove daylight noise and return the magnitude of the noise (stdn)
   and the first gate (down from the top) that contains a bona fide
   signal */

int
LCCleanRay(int * cloudgate,float * stdn,LCBackscatterProfile * profile)
{
  return LCCustomCleanRay(cloudgate, stdn, profile, 16, 4.0);
}


/* Lidar data is often glitched in a way that is difficult to
   characterise physically. This is because we have little idea of
   what Vaisala put in their processing algorithms. The main problem
   is noisy background values, and even betas that go negative. This
   algorithm is a cynical attempt to get rid of as much bad data as
   possible while keeping as much good data as possible. It's a dirty
   job (black art?) but someone's got to do it. */
int
LCCustomCleanRay(int * cloudgate,float * stdn,
		 LCBackscatterProfile * profile,int n,float noisefactor)
{
  int i, j;
  int sumsq=0;
  int thresh, last_gap=0, first_gap=0;
  int accept, sumbeta=0, ngates=profile->ngates;
  int beta;

  /* First calculate the standard deviation of the noise from the top
     n gates We assume it's cloud free - could be a mistake? */
  for (i=ngates-n-1;i++ < ngates-1;
       sumsq += profile->beta[i] * profile->beta[i])
    ;
  *stdn = sqrt( (double) sumsq / (double) n );
  thresh = ((*stdn) * noisefactor);

  /* Now go up through the profile...  We ignore the first gate for
     the moment because it is often empty (why?) */
  accept = 10;
  for (j = 1; j < ngates; ++j) {
    /* Keep a running sum of the betas we have so far encountered */
    if ((beta = profile->beta[j]) > 0) {
      sumbeta += beta;
    }
    else {
      /* Found a gap... */
      if (!first_gap) {
	first_gap = j;
      }

      /* If negative values are found then redistribute them above.
	 Why? Well, negative backscatter coefficients are ridiculous,
	 and it seems like dodgy negative values are often accompanied
	 by dodgy positive values. Anyway, it makes the data look
	 cleaner - I told you it was a black art. */
      if (beta < 0 && j < ngates-1) {
	profile->beta[j+1] += beta;
	profile->beta[j] = 0;
      }

      /* If the positive data since the last gap has not exceeded the
	 acceptance threshold twice then get rid of it all */
      if (accept < 2) {
	for (i = last_gap; i < j; ++i) {
	  profile->beta[i] = 0;
	}
      }
      /* In case any negative value has been missed (such as at the top
	 of the ray, set it to zero */
      profile->beta[i] = 0;
      last_gap = j;
      accept = 0;
      continue;
    }

    /* Positive value found. Our acceptance threshold is the sum of:
       > thresh: based on the standard deviation of the noise measured 
       at the top of the ray;
       > sumbeta/?: term to account for the fact that above large signals
       we should not be measuring small values. This gets rid of some 
       dodgy lines above optically thick cloud.
       > gaussian term: Basically, we want to keep boundary layer aerosol,
       but get rid of isolated low values elsewhere. This term increases
       from 0 at the surface to 3 at infinity. */
    if (beta > (thresh + (sumbeta/500)
	+ (int) (3.0*(1-exp(-1e-7 * profile->range[j] * profile->range[j]))
	))) {
      ++accept;
    }
  }
  /* remove remaining isolated pixels, whatever their value... */
  for (j = 1; j < ngates-1; ++j) {
    if (profile->beta[j-1] < 1 && profile->beta[j+1] < 1) {
      profile->beta[j] = 0;
    }
  }

  /* First gate has not been touched - check it isn't negative */
  if (profile->beta[0] < 0) {
    profile->beta[0] = 0;
  }

  return first_gap;
}
